#pragma once
#include "koopa.h"
#include <string>
using namespace std;

void Visit(const koopa_raw_function_t &func, string& str);
void Visit(const koopa_raw_basic_block_t &bb, string& str);
void Visit(const koopa_raw_value_t &value, string& str);
void Visit(const koopa_raw_return_t &ret, string& str);
void Visit(const koopa_raw_integer_t &integer, string& str);
void Visit(const koopa_raw_binary_t &binary, string& str);

static string value_ret = string();
static int reg_counter = 0;

void Visit(const koopa_raw_slice_t &slice, string& str)
{
    for (size_t i = 0; i < slice.len; ++i)
    {
        auto ptr = slice.buffer[i];
        switch (slice.kind)
        {
        case KOOPA_RSIK_FUNCTION:
            Visit(reinterpret_cast<koopa_raw_function_t>(ptr), str);
            break;
        case KOOPA_RSIK_BASIC_BLOCK:
            Visit(reinterpret_cast<koopa_raw_basic_block_t>(ptr), str);
            break;
        case KOOPA_RSIK_VALUE:
            Visit(reinterpret_cast<koopa_raw_value_t>(ptr), str);
            break;
        default:
            assert(false);
        }
    }
}

void Root_Visit(const koopa_raw_program_t &program, string& str)
{
    str.append("  .text\n");
    str.append("  .global main\n");
    Visit(program.values, str);
    Visit(program.funcs, str);
}

inline bool is_i(const string& str)
{
    if(str[0]=='t'||str[0]=='a'||str[0]=='x') return false;
    return true;
}

inline string get_reg()
{
    if(reg_counter <= 6) return string("t")+to_string(reg_counter++);
    else if (reg_counter <= 14) return string("a")+to_string((reg_counter++)-7);
    else assert(false);
}

void Visit(const koopa_raw_function_t &func, string& str)
{
    str.append(string(func->name).substr(1) + ":\n");
    Visit(func->bbs, str);
}

void Visit(const koopa_raw_basic_block_t &bb, string& str)
{
    const koopa_raw_slice_t &slice = bb->insts;
    for (size_t i = 0; i < slice.len; ++i)
    {
        auto ptr = slice.buffer[i];
        if(reinterpret_cast<koopa_raw_value_t>(ptr)->kind.tag==KOOPA_RVT_RETURN)
        {
            Visit(reinterpret_cast<koopa_raw_value_t>(ptr), str);
            break;
        }        
    }
    // Visit(bb->insts, str);
}

void Visit(const koopa_raw_value_t &value, string& str)
{
    const auto &kind = value->kind;
    switch (kind.tag)
    {
    case KOOPA_RVT_RETURN:
        Visit(kind.data.ret, str);
        break;
    case KOOPA_RVT_INTEGER:
        Visit(kind.data.integer, str);
        break;
    case KOOPA_RVT_BINARY:
        Visit(kind.data.binary, str);
        break;
    default:
        assert(false);
    }
}

void Visit(const koopa_raw_return_t &ret, string& str)
{
    Visit(ret.value, str);
    if(is_i(value_ret))
    {
        str.append("  li    a0, "+value_ret+"\n  ret\n");
    }else{
        str.append("  mv    a0, "+value_ret+"\n  ret\n");
    }
    value_ret = string();
}

void Visit(const koopa_raw_integer_t &integer, string& str)
{
    value_ret = to_string(integer.value);
    if(value_ret == "0") value_ret = "x0";
}

void Visit(const koopa_raw_binary_t &binary, string& str)
{
    Visit(binary.lhs, str);
    string lhs = value_ret;
    if(is_i(lhs)){
        lhs = get_reg();
        str.append("  li    "+lhs+", "+value_ret+"\n");
    }
    value_ret = string();
    Visit(binary.rhs, str);
    string rhs = value_ret;
    value_ret = string();
    
    switch(binary.op){
        case KOOPA_RBO_ADD:
            value_ret = get_reg();
            if (is_i(rhs)){
                str.append("  addi  "+value_ret+", "+lhs+", "+rhs+"\n");
            }else{
                str.append("  add   "+value_ret+", "+lhs+", "+rhs+"\n");
            }
            break;
        case KOOPA_RBO_SUB:         
            if (is_i(rhs)){
                string tmp_rhs = rhs;
                rhs = get_reg();
                str.append("  li    "+rhs+", "+tmp_rhs+"\n");
            }
            value_ret = get_reg();
            str.append("  sub   "+value_ret+", "+lhs+", "+rhs+"\n");
            break;
        case KOOPA_RBO_MUL:
            if (is_i(rhs)){
                string tmp_rhs = rhs;
                rhs = get_reg();
                str.append("  li    "+rhs+", "+tmp_rhs+"\n");
            }
            value_ret = get_reg();
            str.append("  mul   "+value_ret+", "+lhs+", "+rhs+"\n");
            break;
        case KOOPA_RBO_DIV:
            if (is_i(rhs)){
                string tmp_rhs = rhs;
                rhs = get_reg();
                str.append("  li    "+rhs+", "+tmp_rhs+"\n");
            }
            value_ret = get_reg();
            str.append("  div   "+value_ret+", "+lhs+", "+rhs+"\n");
            break;
        case KOOPA_RBO_MOD:
            if (is_i(rhs)){
                string tmp_rhs = rhs;
                rhs = get_reg();
                str.append("  li    "+rhs+", "+tmp_rhs+"\n");
            }
            value_ret = get_reg();
            str.append("  rem   "+value_ret+", "+lhs+", "+rhs+"\n");
            break;
        case KOOPA_RBO_AND:
            break;
        case KOOPA_RBO_OR:
            break;
        case KOOPA_RBO_XOR:
            break;
        case KOOPA_RBO_SHL:
            break;
        case KOOPA_RBO_SHR:
            break;
        case KOOPA_RBO_SAR:
            break;
        case KOOPA_RBO_NOT_EQ:
            value_ret = get_reg();
            if (is_i(rhs)){
                str.append("  xori  "+value_ret+", "+lhs+", "+rhs+"\n");
            }else{
                str.append("  xor   "+value_ret+", "+lhs+", "+rhs+"\n");
            }
            str.append("  snez  "+value_ret+", "+value_ret+"\n");
            break;
        case KOOPA_RBO_EQ:
            value_ret = get_reg();
            if (is_i(rhs)){
                str.append("  xori  "+value_ret+", "+lhs+", "+rhs+"\n");
            }else{
                str.append("  xor   "+value_ret+", "+lhs+", "+rhs+"\n");
            }
            str.append("  seqz  "+value_ret+", "+value_ret+"\n");
            break;
        case KOOPA_RBO_GT:
            break;
        case KOOPA_RBO_LT:
            break;
        case KOOPA_RBO_GE:
            break;
        case KOOPA_RBO_LE:
            break;
    }
}
